# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod, abstractproperty

"""Marmee module."""


class AbstractMarmee(object):
    """Give docstring for AbstractMarmee."""

    __metaclass__ = ABCMeta

    @abstractproperty
    def name(self):
        """Give the name."""
        raise NotImplementedError  # abstract

    @abstractmethod
    def get_name(self):
        """Set the method that should give the name."""
        raise NotImplementedError  # abstract

    @abstractmethod
    def is_marmee(self):
        """Give if it is the implementation."""
        raise NotImplementedError  # abstract

    @abstractproperty
    def inputs(self):
        """Give the inputs of a calculation at certain point."""
        raise NotImplementedError  # abstract

    @abstractmethod
    def set_inputs(self):
        """Set the method that should place the inputs."""
        raise NotImplementedError  # abstract

    @abstractproperty
    def outputs(self):
        """Give the outputs of a calculation at certain point."""
        raise NotImplementedError  # abstract

    @abstractmethod
    def get_outputs(self):
        """Set the method that should give the outputs."""
        raise NotImplementedError  # abstract

    @abstractproperty
    def filters(self):
        """Give the filters of a calculation at certain point."""
        raise NotImplementedError  # abstract

    @abstractmethod
    def set_filters(self):
        """Set the method that should place the filters."""
        raise NotImplementedError  # abstract



# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from marshmallow import Schema, fields, ValidationError
from .base import MarmeeObject
from .filter import FilterSchema
from pystac.models.item import Item
import json


class Input(MarmeeObject):
    def __init__(self, item, reducers):
        self.item = item
        self.reducers = reducers

    def __repr__(self):
        return '<Input(item={self.item.id!r})>'.format(self=self)

    @property
    def dict(self):
        return dict(
            item=self.item.dict,
            reducers=[reducer.dict for reducer in self.reducers]
        )

    @property
    def json(self):
        # return InputSchema().dumps(
        #     self
        # )
        return json.dumps(self.dict)


class InputSchema(Schema):
    item = fields.Method('get_item', deserialize='load_item')
    reducers = fields.Nested(FilterSchema, many=True)

    def get_item(self, obj):
        try:
            isinstance(obj, Item)
            return obj.dict
        except ValidationError as e:
            raise

    def load_item(self, obj):
        return obj.json



# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from .filter import Filter
from .input import Input
from .output import Output
from .rule import Rule, Range

__all__ = ['filter', 'input', 'output', 'rule']



# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from marshmallow import Schema, fields
from .base import MarmeeObject
from .rule import RuleSchema
import json


class Filter(MarmeeObject):
    def __init__(self, name, rules):
        self.name = name
        self.rules = rules

    def __repr__(self):
        return '<Filter(name={self.name!r})>'.format(self=self)

    @property
    def dict(self):
        return dict(
            name=self.name,
            rules=[rule.dict for rule in self.rules]
        )

    @property
    def json(self):
        # return FilterSchema().dumps(
        #     self
        # )
        return json.dumps(self.dict)


class FilterSchema(Schema):
    name = fields.Str()
    rules = fields.Nested(RuleSchema, many=True)



# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from marshmallow import Schema, fields, ValidationError
from .base import MarmeeObject
from pystac.models.item import Item


class Output(MarmeeObject):
    def __init__(self, item):
        self.item = item

    def __repr__(self):
        return '<Output(item={self.item.id!r})>'.format(self=self)

    @property
    def dict(self):
        return dict(
            item=self.item
        )

    @property
    def json(self):
        return OutputSchema().dumps(
            self
        )


class OutputSchema(Schema):
    item = fields.Method('get_item', deserialize='load_item')

    def get_item(self, obj):
        try:
            isinstance(obj, Item)
            return obj.dict
        except ValidationError as e:
            raise

    def load_item(self, obj):
        return obj.json



#!/usr/bin/env python
# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

# -*- coding: utf-8 -*-

"""Tests for `marmee.model` package."""

import pytest
from marshmallow import fields, post_load
from marmee.model.input import (
    Input,
    InputSchema
)
from marmee.model.filter import (
    Filter,
    FilterSchema,
    RuleSchema
)
from marshmallow import pprint


class TestInput(object):
    """docstring for testing marmee.model.input ."""

    @pytest.fixture
    def input(self):

        class FilterSchemaInstance(FilterSchema):

            def __init__(self, name, rules):
                self.name = name
                self.rules = rules

            @post_load
            def make_filter(self, data):
                return Filter(**data)

        # first_arg = ArgumentSchemaInstance().load(
        #     positional=self.positional,
        #     identifier=self.identifier,
        #     values=self.values #  fields.List.load(["first","second"])
        # )
        # first_arg = FilterSchemaInstance(
        #     "first",
        #     ["first", "second"]
        # )
        # second_arg = FilterSchemaInstance(
        #     "second",
        #     ["first", "second"]
        # )

        # pprint(first_arg)

        # return InputSchema().load(
        #     process="firttest",
        #     arguments={
        #         first_arg,
        #         second_arg
        #     }
        # )

    def test_mandatory(self):
        print(self)
        try:
            assert True
            # assert self.input().process
            # assert self.process
            # assert self.arguments
        except Exception as e:
            raise



# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from marshmallow import fields, Schema, post_load
from .base import MarmeeObject
from marshmallow_oneofschema import OneOfSchema
import json


class SpatialRule(MarmeeObject):
    def __init__(self, extent):
        self.extent = extent

    @property
    def dict(self):
        return dict(
            extent=self.extent
        )

    @property
    def json(self):
        return SpatialRuleSchema().dumps(
            self
        )


class TemporalRule(MarmeeObject):
    def __init__(self, daterange):
        self.daterange = daterange

    @property
    def dict(self):
        return dict(
            daterange=self.daterange
        )

    @property
    def json(self):
        return TemporalRuleSchema().dumps(
            self
        )


class SpatialRuleSchema(Schema):
    extent = fields.Dict(required=True)

    @post_load
    def make_extent(self, data):
        return SpatialRule(**data)


class Range(MarmeeObject):
    def __init__(self, from_date, to_date):
        self.from_date = from_date
        self.to_date = to_date

    @property
    def dict(self):
        return dict(
            from_date=self.from_date.strftime("%Y-%m-%d"),
            to_date=self.to_date.strftime("%Y-%m-%d")
        )

    @property
    def json(self):
        return RangeSchema().dumps(
            self
        )


class RangeSchema(Schema):
    from_date = fields.DateTime()
    to_date = fields.DateTime()


class TemporalRuleSchema(RangeSchema):
    daterange = fields.Dict(required=True)

    @post_load
    def make_daterange(self, data):
        return TemporalRule(**data)


# class Extent(MarmeeObject):
#     def __init__(self, t_rule, s_rule):
#         # self = TemporalRule(t_rule.dict), SpatialRule(s_rule.dict)
#         self = ExtentSchema(many=True).dump(
#             [t_rule, s_rule]
#         )

#     @property
#     def dict(self):
#         return json.loads(self)

#     @property
#     def json(self):
#         # return ExtentSchema().dump(
#         #     self, many=True
#         # )
#         return json.dumps(self.dict)


class ExtentSchema(OneOfSchema):
    type_field = 'type'
    type_field_remove = False
    type_schemas = {
        'spatial': SpatialRuleSchema,
        'temporal': TemporalRuleSchema,
    }

    def get_obj_type(self, obj):
        if isinstance(obj, SpatialRule):
            return 'spatial'
        elif isinstance(obj, TemporalRule):
            return 'temporal'
        else:
            raise Exception('Unknown object type: %s' % obj.__class__.__name__)


class Rule(MarmeeObject):
    def __init__(self, identifier, rule):
        self.identifier = identifier
        self.rule = rule

    @property
    def dict(self):
        return dict(
            identifier=self.identifier,
            # rule=[r.dict for r in self.rule]
            rule=self.rule
        )

    @property
    def json(self):
        # return RuleSchema().dumps(
        #     self
        # )
        return json.dumps(self.dict)


class RuleSchema(Schema):
    identifier = fields.Str()
    rule = fields.Nested(ExtentSchema)



# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

# -*- coding: utf-8 -*-

"""Top-level package for marmee."""

__author__ = """Francesco Bartoli"""
__email__ = 'francesco.bartoli@geobeyond.it'
__version__ = '0.1.7'



from marshmallow import (
    Schema,
    fields
)


class MarmeeObject(object):

    def __init__(self, marmee):
        self.marmee = marmee

    @property
    def dict(self):
        return self.marmee

    @property
    def json(self):
        return MarmeeObjectSchema().dumps(
            self
        )


class MarmeeObjectSchema(Schema):

    marmee = fields.Dict()



#!/usr/bin/env python
# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

# -*- coding: utf-8 -*-

"""Tests for `marmee` package."""

import pytest
from marmee.marmee import Marmee

@pytest.fixture
def set_inputs():
    inputs = {}
    return inputs


class TestMarmee(object):
    """docstring for testing marmee."""

    cls = Marmee

    def test_is_marmee(self):
        try:
            self.cls().is_marmee()
        except (NotImplementedError, TypeError):
            assert True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    # TODO: Put package requirements here
    'pendulum',
    'earthengine-api',
    'marshmallow==3.0.0b9',
    'marshmallow-oneofschema>=1.0.6',
    'stacpy'
]

deps_links = [
    'git+https://github.com/francbartoli/marshmallow-oneofschema.git@1.0.6#egg=marshmallow-oneofschema-1.0.6'
]

setup_requirements = [
    'pytest-runner',
    # TODO(francbartoli): Put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'pytest',
    # TODO: Put package test requirements here
]

setup(
    name='marmee',
    version='0.1.7',
    description="Execute safely Google Earth Engine algorithm through\
a tunnel like a marmot",
    long_description=readme + '\n\n' + history,
    author="Francesco Bartoli",
    author_email='francesco.bartoli@geobeyond.it',
    url='https://github.com/francbartoli/marmee',
    packages=find_packages(include=['marmee']),
    include_package_data=True,
    install_requires=requirements,
    dependency_links=deps_links,
    license="MIT license",
    zip_safe=False,
    keywords='marmee',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



#!/usr/bin/env python
# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

# -*- coding: utf-8 -*-

"""Tests for `marmee` package."""

import pytest
from marmee.abstract_marmee import AbstractMarmee


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


class TestAbstractMarmee(object):
    """docstring for testing marmee."""

    cls = AbstractMarmee

    def test_is_marmee(self):
        try:
            self.cls().is_marmee()
        except (NotImplementedError, TypeError):
            assert True

    def test_name(self):
        try:
            self.cls().name
        except (NotImplementedError, TypeError):
            assert True

    def test_get_name(self):
        try:
            self.cls().get_name()
        except (NotImplementedError, TypeError):
            assert True

    def test_inputs(self):
        try:
            self.cls().inputs
        except (NotImplementedError, TypeError):
            assert True

    def test_set_inputs(self):
        try:
            self.cls().set_inputs()
        except (NotImplementedError, TypeError):
            assert True

    def test_results(self):
        try:
            self.cls().outputs
        except (NotImplementedError, TypeError):
            assert True

    def test_get_results(self):
        try:
            self.cls().get_outputs()
        except (NotImplementedError, TypeError):
            assert True

    def test_filters(self):
        try:
            self.cls().filters
        except (NotImplementedError, TypeError):
            assert True

    def test_set_filters(self):
        try:
            self.cls().set_filters()
        except (NotImplementedError, TypeError):
            assert True



# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT
# -*- coding: utf-8 -*-
from .abstract_marmee import AbstractMarmee
from marshmallow import fields, Schema, ValidationError
from .model.base import MarmeeObject
from .model.input import InputSchema
from .model.output import OutputSchema
from .model.filter import FilterSchema
import json


class Marmee(AbstractMarmee, MarmeeObject):

    def __init__(self, name, inputs, filters, outputs):

        self._name = name
        self._inputs = inputs
        self._filters = filters
        self._outputs = []

    @property
    def name(self):
        return self._name

    @name.setter
    def set_name(self, val):
        try:
            self.name = val
        except ValidationError as ve:
            raise

    @property
    def inputs(self):
        return self._inputs

    @inputs.setter
    def set_inputs(self, *args):
        for arg in args:
            try:
                self._inputs.append(arg)
            except ValidationError as e:
                raise

    @property
    def filters(self):
        return self._filters

    @filters.setter
    def set_filters(self, *args):
        for arg in args:
            try:
                self._filters.append(arg)
            except ValidationError as e:
                raise

    @property
    def outputs(self):
        return self._outputs

    @outputs.getter
    def get_outputs(self):
        return self.outputs

    @outputs.setter
    def set_outputs(self, *args):
        for arg in args:
            try:
                self._outputs.append(arg)
            except ValidationError as e:
                raise

    def get_name(self):
        return self.name

    def is_marmee(self):
        pass

    @property
    def dict(self):
        return dict(
            name=self.name,
            inputs=[input.dict for input in self.inputs],
            filters=[filter.dict for filter in self.filters],
            outputs=[output.dict for output in self.outputs],
        )

    @property
    def json(self):
        # return MarmeeSchema().dumps(
        #     self
        # )
        return json.dumps(self.dict)


class MarmeeSchema(Schema):
    name = fields.Str()
    inputs = fields.Nested(InputSchema, many=True)
    filters = fields.Nested(FilterSchema, many=True)
    outputs = fields.Nested(OutputSchema, many=True)



# Copyright (c) 2018 Francesco Bartoli
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from marshmallow import ValidationError
from pystac.models.item import Item
from pystac.models.asset import Asset
from pystac.models.link import Link
from pystac.models.collection import Collection
from pystac.models.properties import Properties
from ee import EEException, data, Image, ImageCollection
import pendulum
import dask
from dask.delayed import delayed
import json
import os

TOKEN_TYPE = [
    ('Image', 'Item'),
    ('ImageCollection', 'Collection')
]


class Stac(object):
    def __init__(self, gee):
        self.gee = gee

    @property
    def type(self):
        if TOKEN_TYPE[0][0] == self._get_info(
        ).get('type'):
            return TOKEN_TYPE[0][1]
        elif TOKEN_TYPE[1][0] == self._get_info(
        ).get('type'):
            return TOKEN_TYPE[1][1]

    def parse(self):
        """Parse an asset from Earth Engine to STAC item

        Raises:
            ValueError -- If asset is an ImageCollection

        Returns:
            Item -- STAC feature of the Google Earth Engine Asset
            Collection -- STAC collection of the Google Earth Engine Asset
        """

        if self.type == TOKEN_TYPE[0][1]:
            try:
                return Item(
                    item_id=self._link(None, None)[1],
                    links=self._link(None, None)[0],
                    assets=self._asset(None),
                    properties=self._properties(None)[0],
                    geometry=self._properties(None)[2]
                )
            except ValidationError as e:
                raise
        elif self.type == TOKEN_TYPE[1][1]:
            try:
                # parallelize item computation
                items = [self._features_iterator(
                    feature['id'],
                    self._link(feature, data.ASSET_TYPE_IMAGE_COLL)[0],
                    self._asset(
                        feature['properties']['system:index']
                    ),
                    self._properties(feature)[0],
                    self._properties(feature)[2]
                ) for feature in self._get_full_info()['features']]
                res_list = dask.compute(items)[0]

                return Collection(
                    features=res_list
                )
            except ValidationError as e:
                raise
        else:
            raise TypeError("Unrecognized Stac type found.")

    def _get_info(self):
        return data.getInfo(self.gee)

    def _get_full_info(self):
        if self.type in TOKEN_TYPE[0][1]:
            return data.getValue(
                {'json': Image(self.gee).serialize()}
            )
        elif self.type in TOKEN_TYPE[1][1]:
            return data.getValue(
                {'json': ImageCollection(self.gee).serialize()}
            )

    def _properties(self, gson):
        if not gson:
            gson = self._get_info()
        if 'properties' in gson.keys():
            extprops = gson.get('properties')
            provider = license = dtime = index = ""
            footprint = {}
            for key in extprops.keys():
                if isinstance(json.dumps(extprops[key]), str):
                    if key == "provider":
                        provider = extprops[key]
                    if key == "license":
                        license = extprops[key]
                    if key == "system:time_start":
                        dtime = pendulum.from_timestamp(
                            extprops[key] / 1000
                        ).to_rfc3339_string()
                    if dtime and key == "system:time_end":
                        dtime = dtime + "/" + pendulum.from_timestamp(
                            extprops[key] / 1000
                        ).to_rfc3339_string()
                    if key == "system:index":
                        index = extprops[key]
                if isinstance(extprops[key], dict):
                    if key == "system:footprint":
                        footprint = extprops[key]
            try:
                prop = Properties(
                    datetime=dtime,
                    asset_license=license,
                    provider=provider,
                    ext_properties=extprops
                )
                return (prop, index, footprint,)
            except ValidationError as e:
                raise

    def _asset(self, gson):
        if not gson:
            gson = self._properties(None)[1]
        href = name = gson
        try:
            asset = Asset(href=href, name=name)
            return {asset.name: asset.dict}
        except ValidationError as e:
            raise

    def _link(self, gson, rel_asset):
        lang = "EN"
        linktype = ""
        if not gson:
            gson = self._get_info()
        if 'id' in gson:
            ident = gson.get('id')
            href = os.path.dirname(ident)
            if not rel_asset:
                if data.getInfo(
                    href
                ).get('type') in data.ASSET_TYPE_IMAGE_COLL:
                    rel = data.ASSET_TYPE_IMAGE_COLL
                elif data.getInfo(
                    href
                ).get('type') in data.ASSET_TYPE_FOLDER:
                    rel = data.ASSET_TYPE_FOLDER
            else:
                rel = rel_asset
            try:
                return (
                    [Link(
                        href=href,
                        rel=rel,
                        hreflang=lang,
                        link_type=linktype
                    )],
                    ident,
                )
            except ValidationError as e:
                raise
        else:
            raise EEException("Asset has to be with unique id property")

    @delayed
    def _features_iterator(ft_id, link, asset, props, geom):

        return Item(
            item_id=ft_id,
            links=link,
            assets=asset,
            properties=props,
            geometry=geom
        )
